@@ -16,22 +16,25 @@ try:
     import paramiko
 except ImportError:
     raise RuntimeError(
-        (
-            "You must install paramiko package (pip install paramiko) "
-            "to use the paramiko backend"
-        )
-    )
+        "You must install paramiko package (pip install paramiko) "
+        "to use the paramiko backend"
+    ) from None
 
+import functools
+from typing import Any, Optional
+
+import paramiko.pkey
 import paramiko.ssh_exception
 
 from testinfra.backend import base
-from testinfra.utils import cached_property
 
 
 class IgnorePolicy(paramiko.MissingHostKeyPolicy):
     """Policy for ignoring missing host key."""
 
-    def missing_host_key(self, client, hostname, key):
+    def missing_host_key(
+        self, client: paramiko.SSHClient, hostname: str, key: paramiko.pkey.PKey
+    ) -> None:
         pass
 
 
@@ -40,12 +43,12 @@ class ParamikoBackend(base.BaseBackend):
 
     def __init__(
         self,
-        hostspec,
-        ssh_config=None,
-        ssh_identity_file=None,
-        timeout=10,
-        *args,
-        **kwargs
+        hostspec: str,
+        ssh_config: Optional[str] = None,
+        ssh_identity_file: Optional[str] = None,
+        timeout: int = 10,
+        *args: Any,
+        **kwargs: Any,
     ):
         self.host = self.parse_hostspec(hostspec)
         self.ssh_config = ssh_config
@@ -54,7 +57,13 @@ class ParamikoBackend(base.BaseBackend):
         self.timeout = int(timeout)
         super().__init__(self.host.name, *args, **kwargs)
 
-    def _load_ssh_config(self, client, cfg, ssh_config):
+    def _load_ssh_config(
+        self,
+        client: paramiko.SSHClient,
+        cfg: dict[str, Any],
+        ssh_config: paramiko.SSHConfig,
+        ssh_config_dir: str = "~/.ssh",
+    ) -> None:
         for key, value in ssh_config.lookup(self.host.name).items():
             if key == "hostname":
                 cfg[key] = value
@@ -74,9 +83,17 @@ class ParamikoBackend(base.BaseBackend):
                 cfg["gss_kex"] = value == "yes"
             elif key == "proxycommand":
                 cfg["sock"] = paramiko.ProxyCommand(value)
+            elif key == "include":
+                new_config_path = os.path.join(
+                    os.path.expanduser(ssh_config_dir), value
+                )
+                with open(new_config_path) as f:
+                    new_ssh_config = paramiko.SSHConfig()
+                    new_ssh_config.parse(f)
+                    self._load_ssh_config(client, cfg, new_ssh_config, ssh_config_dir)
 
-    @cached_property
-    def client(self):
+    @functools.cached_property
+    def client(self) -> paramiko.SSHClient:
         client = paramiko.SSHClient()
         client.set_missing_host_key_policy(paramiko.WarningPolicy())
         cfg = {
@@ -87,48 +104,56 @@ class ParamikoBackend(base.BaseBackend):
             "password": self.host.password,
         }
         if self.ssh_config:
+            ssh_config_dir = os.path.dirname(self.ssh_config)
+
             with open(self.ssh_config) as f:
                 ssh_config = paramiko.SSHConfig()
                 ssh_config.parse(f)
-                self._load_ssh_config(client, cfg, ssh_config)
+                self._load_ssh_config(client, cfg, ssh_config, ssh_config_dir)
         else:
             # fallback reading ~/.ssh/config
             default_ssh_config = os.path.join(os.path.expanduser("~"), ".ssh", "config")
+            ssh_config_dir = os.path.dirname(default_ssh_config)
+
             try:
                 with open(default_ssh_config) as f:
                     ssh_config = paramiko.SSHConfig()
                     ssh_config.parse(f)
-            except IOError:
+            except OSError:
                 pass
             else:
-                self._load_ssh_config(client, cfg, ssh_config)
+                self._load_ssh_config(client, cfg, ssh_config, ssh_config_dir)
 
         if self.ssh_identity_file:
             cfg["key_filename"] = self.ssh_identity_file
-        client.connect(**cfg)
+        client.connect(**cfg)  # type: ignore[arg-type]
         return client
 
-    def _exec_command(self, command):
-        chan = self.client.get_transport().open_session()
+    def _exec_command(self, command: bytes) -> tuple[int, bytes, bytes]:
+        transport = self.client.get_transport()
+        assert transport is not None
+        chan = transport.open_session()
         if self.get_pty:
             chan.get_pty()
         chan.exec_command(command)
-        rc = chan.recv_exit_status()
         stdout = b"".join(chan.makefile("rb"))
         stderr = b"".join(chan.makefile_stderr("rb"))
+        rc = chan.recv_exit_status()
         return rc, stdout, stderr
 
-    def run(self, command, *args, **kwargs):
+    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
         command = self.get_command(command, *args)
-        command = self.encode(command)
+        cmd = self.encode(command)
         try:
-            rc, stdout, stderr = self._exec_command(command)
-        except paramiko.ssh_exception.SSHException:
-            if not self.client.get_transport().is_active():
+            rc, stdout, stderr = self._exec_command(cmd)
+        except (paramiko.ssh_exception.SSHException, ConnectionResetError):
+            transport = self.client.get_transport()
+            assert transport is not None
+            if not transport.is_active():
                 # try to reinit connection (once)
                 del self.client
-                rc, stdout, stderr = self._exec_command(command)
+                rc, stdout, stderr = self._exec_command(cmd)
             else:
                 raise
 
-        return self.result(rc, command, stdout, stderr)
+        return self.result(rc, cmd, stdout, stderr)
